# frozen_string_literal: true

require 'bunny'
require 'warren/fox'
require 'warren/subscription'
require 'warren/delay_exchange'

module Warren
  # A Den is in charge of creating a Fox from a consumer configuration
  # It handles the registration of dead-letter queues, and configuration of
  # {Warren::Subscription subscriptions} and
  # {Warren::DelayExchange delay exchanges}
  class Den
    # The number of simultaneous workers generated by default
    DEFAULT_WORKER_COUNT = 3

    #
    # Create a {Warren::Fox} work pool.
    # @param app_name [String] The name of the application. Corresponds to the
    #                          subscriptions config in `config/warren.yml`
    # @param config [Warren::Config::Consumers] A configuration object, loaded from `config/warren.yml` by default
    # @param adaptor [#recovered?,#handle,#env] An adaptor to handle framework specifics
    def initialize(app_name, config, adaptor:)
      @app_name = app_name
      @config = config
      @fox = nil
      @adaptor = adaptor
    end

    def fox
      @fox ||= spawn_fox
    end

    #
    # Ensures the dead_letter queues and exchanges are registered.
    #
    # @return [Void]
    def register_dead_letter_queues
      config = dead_letter_config
      return unless config

      Warren.handler.with_channel do |channel|
        subscription = Warren::Subscription.new(channel: channel, config: config)
        subscription.activate!
      end
    end

    private

    def consumer_config
      @config.consumer(@app_name)
    end

    #
    # Spawn a new fox
    #
    # @return [Warren::Fox]
    def spawn_fox
      # We don't use with_channel as our consumer persists outside the block,
      # and while we *can* share channels between consumers it results in them
      # sharing the same worker pool. This process lets us control workers on
      # a per-queue basis. Currently that just means one worker per consumer.
      channel = Warren.handler.new_channel(worker_count: worker_count)
      subscription = Warren::Subscription.new(channel: channel, config: queue_config)
      delay = Warren::DelayExchange.new(channel: channel, config: delay_config)
      Warren::Fox.new(name: @app_name,
                      subscription: subscription,
                      adaptor: @adaptor,
                      subscribed_class: subscribed_class,
                      delayed: delay)
    end

    def worker_count
      consumer_config.fetch('worker_count', DEFAULT_WORKER_COUNT)
    end

    def queue_config
      consumer_config.fetch('queue')
    end

    def dead_letter_config
      consumer_config.fetch('dead_letters')
    end

    def delay_config
      consumer_config.fetch('delay', nil)
    end

    def subscribed_class
      Object.const_get(consumer_config.fetch('subscribed_class'))
    end
  end
end
